@@ -15,6 +15,7 @@ from functools import lru_cache
 import os.path as op
 import nibabel as nib
 from fsl.data import image
+from sympy.utilities.autowrap import ufuncify
 
 
 def default_dependencies():
@@ -37,7 +38,7 @@ def default_dependencies():
             [True, True, False, False],   # f_myelin
             [True, False, False, False],  # log_g_myelin
             [False, True, True, False],   # non_myelin
-            [True, True, False, False],   # amplitude_se
+            [True, True, True, False],   # amplitude_se
             [False, False, True, True],   # R2_attenuation
             [True, False, True, False],   # R2
             [True, False, True, False],   # R2i
@@ -434,17 +435,15 @@ class Model(object):
     """
     Model to fit to the DIPPI data
     """
-    def __init__(self, shells: Sequence[Shell], T2=False, dual_phase=False, only_phase=False, with_intercept=False):
+    def __init__(self, scanner, T2=False, dual_phase=False, only_phase=False, with_intercept=False):
         """
         Creates a new DIPPI model
 
-        :param shells: Describes the acquisition parameters
         :param T2: if True fit R2 and R2* instead of fitting each amplitude independently
         :param dual_phase: if True fit unmyelinated and myelinated axons as separate compartments
         :param only_phase: Ignore the biophysical model and just fit the phase difference between readouts for every dyad
         :param with_intercept: Adds a phase offset at t_phase=0 to the DIPPI model
         """
-        self.shells = shells
         self.T2 = T2
         if dual_phase and only_phase:
             raise ValueError("dual_phase and only_phase are conflicting model definitions")
@@ -458,42 +457,49 @@ class Model(object):
         intercept, movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase = sym.symbols(
             'intercept, movement_phase, eddy_phase, non_myelin, log_g, log_g_myelin, f_myelin, dphase'
         )
+        bval, readout1, readout2, t_phase = sym.symbols(
+            'bval, readout1, readout2, t_phase'
+        )
+        b0_dir = sym.symbols('b0_dir:3')
         bvec = sym.symbols('bvec:3')
 
-        self.magnitude_equations = []
-        self.signal_equations = []
-        for shell in self.shells:
-            profile = sym.exp(-shell.bval * width * alignment(phi, theta, bvec) ** 2)
-            magnitude_se = amplitude_se * profile
-            if T2:
-                R2_attenuation = sym.exp(-(shell.readout2 - shell.readout1) * R2 - shell.t_phase * R2i)
-            magnitude_ase = R2_attenuation * amplitude_se * profile
-            self.magnitude_equations.append((magnitude_se, magnitude_ase))
+        profile = sym.exp(-bval * width * alignment(phi, theta, bvec) ** 2)
+        magnitude_se = amplitude_se * profile
+        if T2:
+            R2_attenuation = sym.exp(-(readout2 - readout1) * R2 - t_phase * R2i)
+        magnitude_ase = R2_attenuation * amplitude_se * profile
+        self.magnitude_equations = (magnitude_se, magnitude_ase)
 
-            signal_se = magnitude_se * sym.exp(1j * movement_phase)
+        signal_se = magnitude_se * sym.exp(1j * movement_phase)
 
-            confound_phase = movement_phase + (non_myelin * shell.t_phase) + eddy_phase
-            if with_intercept:
-                confound_phase = confound_phase + intercept
-            if dual_phase:
-                angle = myelin_phase(shell, phi, theta, log_g_myelin)
-                norm_myelin = sym.sqrt(((1 - f_myelin) + f_myelin * sym.cos(angle)) ** 2 + (f_myelin * sym.sin(angle)) ** 2)
-                myelin_part = (
-                    f_myelin * sym.exp(1j * angle) +
-                    (1 - f_myelin)
-                ) / norm_myelin
-                signal_ase = magnitude_ase * sym.exp(1j * confound_phase) * myelin_part
-            elif only_phase:
-                signal_ase = magnitude_ase * sym.exp(1j * (movement_phase + dphase + eddy_phase))
-            else:
-                signal_ase = magnitude_ase * sym.exp(
-                    1j * (confound_phase + myelin_phase(shell, phi, theta, log_g))
-                )
-            self.signal_equations.append((signal_se, signal_ase))
+        confound_phase = movement_phase + (non_myelin * t_phase) + eddy_phase
+        if with_intercept:
+            confound_phase = confound_phase + intercept
+        if dual_phase:
+            angle = myelin_phase(scanner, b0_dir, t_phase, phi, theta, log_g_myelin)
+            norm_myelin = sym.sqrt(((1 - f_myelin) + f_myelin * sym.cos(angle)) ** 2 + (f_myelin * sym.sin(angle)) ** 2)
+            myelin_part = (
+                f_myelin * sym.exp(1j * angle) +
+                (1 - f_myelin)
+            ) / norm_myelin
+            signal_ase = magnitude_ase * sym.exp(1j * confound_phase) * myelin_part
+        elif only_phase:
+            signal_ase = magnitude_ase * sym.exp(1j * (movement_phase + dphase + eddy_phase))
+        else:
+            signal_ase = magnitude_ase * sym.exp(
+                1j * (confound_phase + myelin_phase(scanner, b0_dir, t_phase, phi, theta, log_g))
+            )
+        self.signal_equations = (signal_se, signal_ase)
 
-        self.parameter_names = tuple(sorted(str(v) for v in signal_ase.free_symbols.union(signal_se.free_symbols)))
+        self.parameters = tuple(sorted((v for v in signal_ase.free_symbols.union(signal_se.free_symbols)), key=str))
+        self.parameter_names = tuple(str(p) for p in self.parameters)
+        self._derivative_equations = {}
+        self._funcs = {}
 
-    def derivative(self, name):
+    def get_equations(self, only_mag=False):
+        return self.magnitude_equations if only_mag else self.signal_equations
+
+    def derivative(self, name, only_mag=True):
         """
         Computes the derivative with respect to a given parameter
 
@@ -502,19 +508,48 @@ class Model(object):
         :param name: name of one of the parameters
         :return: sympy equations with derivative as a list with the equation for each shell and both readouts
         """
-        return [[sym.diff(f, name) for f in fs] for fs in self.signal_equations]
+        index = (name, only_mag)
+        if index not in self._derivative_equations:
+            self._derivative_equations[index] = tuple(sym.simplify(sym.diff(f, name)) for f in self.get_equations(only_mag))
+        return self._derivative_equations[index]
 
-    def funcs(self, derivative=None, only_mag=False):
+    def get_func(self, derivatives=(), only_mag=False):
         """
         Return the model equations as numpy functions
 
-        :param derivative: set to a name to return the derivative with respect to that parameter
+        :param derivatives: list of the derivatives of interest
         :param only_mag: if True only compute the magnitude rather than the complex signal
         :return: for each shell return a tuple with functions for the first and second readout signal (or its derivatives)
         """
-        equations = self.magnitude_equations if only_mag else self.signal_equations
-        return [tuple(_lambdify_equation_cached(self.parameter_names, readout, derivative) for readout in shell)
-                for shell in equations]
+        index = (tuple(derivatives), only_mag)
+        if index not in self._funcs:
+            equations = self.get_equations(only_mag)
+
+            all_equations = [*equations]
+            for name in derivatives:
+                all_equations.extend(self.derivative(name, only_mag))
+
+            intermediates_sym, finals_sym = sym.cse(all_equations)
+            params = list(self.parameter_names)
+            intermediates_num, finals_num = [], {}
+
+            for name, sym_fun in intermediates_sym:
+                intermediates_num.append(sym.lambdify(params, sym_fun))
+                params.append(str(name))
+            
+            for idx_name, name in enumerate(('signal', ) + tuple(derivatives)):
+                finals_num[str(name)] = (
+                    sym.lambdify(params, finals_sym[idx_name * 2]),
+                    sym.lambdify(params, finals_sym[idx_name * 2 + 1]),
+                )
+
+            def func(params):
+                p_internal = list(params)
+                for func in intermediates_num:
+                    p_internal.append(func(*p_internal))
+                return {name: (funcs[0](*p_internal), funcs[1](*p_internal)) for name, funcs in finals_num.items()}
+            self._funcs[index] = func
+        return self._funcs[index]
 
     def func_args(self, parameters: Parameters, to_vary=()):
         """
@@ -522,7 +557,11 @@ class Model(object):
 
         :param parameters: current best estimate of the parameters
         :param to_vary: names of the parameters that should not be filled in yet (None placeholder)
-        :return: for each shell a list of arguments (which are the same for the first and second readout)
+        :return: nested lists with model parameters (they are the same for the first and second readout)
+
+            1. outer list loops over shells
+            2. middle list loops over fibre orientations (length of 2)
+            3. inner list loops over individual parameters in the order of `self.parameter_names`
         """
         res: List[List[List[Optional[np.ndarray]]]] = []
         for shell, sparams in parameters:
@@ -541,6 +580,11 @@ class Model(object):
                     elif name.startswith('bvec'):
                         idx = int(name[-1])
                         value = shell.gradients[:, idx]
+                    elif name.startswith('b0_dir'):
+                        idx = int(name[-1])
+                        value = shell.b0_dir[idx]
+                    elif name in ('bval', 'readout1', 'readout2', 't_phase'):
+                        value = getattr(shell, name)
                     elif name in to_vary:
                         value = None
                     else:
@@ -559,18 +603,19 @@ class Model(object):
         :return: a list with for each shell the signal during the first and second readout
         """
         res: List[List[Any]] = []
-        for shell, all_params, all_funcs in zip(parameters.shells, self.func_args(parameters), self.funcs(only_mag=only_mag)):
-            res.append([])
-            for readout_func in all_funcs:
-                if split_fibres:
-                    res[-1].append([])
-                else:
-                    res[-1].append(np.zeros(shell.ngradients, dtype=float if only_mag else complex))
-                for fibre_param in all_params:
+        func = self.get_func(only_mag=only_mag)
+        for shell, all_params in zip(parameters.shells, self.func_args(parameters)):
+            if split_fibres:
+                res.append([[] for _ in range(2)])
+            else:
+                res.append([np.zeros(shell.ngradients, dtype=float if only_mag else complex) for _ in range(2)])
+            for fibre_param in all_params:
+                signal = func(fibre_param)['signal']
+                for idx_readout in range(2):
                     if split_fibres:
-                        res[-1][-1].append(readout_func(*fibre_param))
+                        res[-1][idx_readout].append(signal[idx_readout])
                     else:
-                        res[-1][-1] += readout_func(*fibre_param)
+                        res[-1][idx_readout] += signal[idx_readout]
         return res
 
     def cost(self, parameters: Parameters, data, only_mag=False):
@@ -608,11 +653,10 @@ class Model(object):
             - function that turns array of best-fit parameter values back into a Parameter object
             - array with initial parameter values
         """
-        all_functions = self.funcs(only_mag=only_mag)
-        all_derivatives = {name: self.funcs(name, only_mag=only_mag) for name in tofit}
+        toderive = list(tofit)
         if 'eddy_odd' in tofit or 'eddy_even' in tofit:
-            all_derivatives['eddy_phase'] = self.funcs('eddy_phase', only_mag=only_mag)
-        array_params = self.func_args(parameters, to_vary=tofit)
+            toderive.append('eddy_phase')
+        all_functions = self.get_func(toderive, only_mag=only_mag)
         struct_params = parameters.copy()
         to_change, map_derivative = struct_params.to_change(tofit)
         init = np.concatenate([[]] + [a.flatten() for a in to_change])
@@ -654,81 +698,10 @@ class Model(object):
             new_params = np.asanyarray(new_params)
             write_struct_params(new_params)
 
-            ct = 0.
-            dp = np.zeros(new_params.size)
-            for idx_shell, (shell, sparams) in enumerate(struct_params):
-                if struct_params.eddy_lmax > 0:
-                    if 'eddy_odd' in tofit:
-                        mat_eddy_odd = shell.spherical_harmonics(struct_params.eddy_lmax, odd=True)
-                    if 'eddy_even' in tofit:
-                        mat_eddy_even = shell.spherical_harmonics(struct_params.eddy_lmax, odd=False)
-                else:
-                    mat_eddy_odd = np.zeros((shell.ngradients, 0))
-                    mat_eddy_even = mat_eddy_odd
-                fibre_params = []
-                for idx_fibre in range(struct_params.nfibres):
-                    param_array = list(array_params[idx_shell][idx_fibre])
-                    for name in tofit:
-                        if name.startswith('eddy_'):
-                            idx = self.parameter_names.index('eddy_phase')
-                            mat = mat_eddy_odd if name == 'eddy_odd' else mat_eddy_even
-                            param_array[idx] = param_array[idx] + mat.dot(sparams[idx_fibre][name])
-                        else:
-                            idx = self.parameter_names.index(name)
-                            param_array[idx] = sparams[idx_fibre][name]
-                    fibre_params.append(param_array)
-
-                for idx_readout in range(2):
-                    signal = np.zeros(shell.ngradients, dtype=float if only_mag else complex)
-                    for param_array in fibre_params:
-                        signal += all_functions[idx_shell][idx_readout](*param_array)
-                    observed = data[idx_shell][idx_readout]
-                    if only_mag or not np.iscomplex(observed).any():
-                        ct_part = ((np.real(signal) - abs(observed)) ** 2).sum()
-                        der_signal = 2 * (np.real(signal) - abs(observed))
-                    else:
-                        ct_part = (abs(signal - observed) ** 2).sum()
-                        der_signal = np.conj(2 * (signal - observed))
-                    ct += ct_part
-
-                    for idx_fibre, param_array in enumerate(fibre_params):
-                        if 'eddy_odd' in tofit or 'eddy_even' in tofit:
-                            dphase = all_derivatives['eddy_phase'][idx_shell][idx_readout](*param_array) * der_signal
-                        for name in tofit:
-                            if name.startswith('eddy_') and struct_params.eddy_lmax == 0:
-                                der_param_seq = np.zeros((), dtype=complex)
-                            elif name == 'eddy_odd':
-                                der_param_seq = shell.spherical_harmonics(struct_params.eddy_lmax, odd=True).T.dot(dphase.real)
-                            elif name == 'eddy_even':
-                                der_param_seq = shell.spherical_harmonics(struct_params.eddy_lmax, odd=False).T.dot(dphase.real)
-                            else:
-                                der_param_seq = all_derivatives[name][idx_shell][idx_readout](*param_array)
-                            if name == 'movement_phase':
-                                der_param = der_param_seq * der_signal
-                            elif name.startswith('eddy_'):
-                                der_param = der_param_seq
-                            elif np.asarray(der_param_seq).ndim == 0:
-                                der_param = der_param_seq * der_signal.sum()
-                            else:
-                                der_param = np.dot(der_param_seq, der_signal)
-                            idx_start = map_derivative[(name, idx_shell, idx_fibre)]
-                            dp[edges[idx_start]:edges[idx_start+1]] += der_param.real
-            return ct, dp
+            return _cost_helper(self, new_params, struct_params, data, all_functions, only_mag, tofit, map_derivative, edges)
         return cost, convert_params, init
 
 
-@lru_cache(maxsize=None)
-def _lambdify_equation_cached(params, equation, name=None):
-    """
-    Cached function to convert sympy equations into numpy functions
-
-    Used internally by `Model.funcs`.
-    """
-    if name is not None:
-        equation = sym.diff(equation, name)
-    return sym.lambdify(params, equation, 'numpy')
-
-
 def alignment(phi, theta, bvec, sympy=True):
     """
     Computes the dot-product between the fibre alignment and the gradient orientation
@@ -747,7 +720,7 @@ def alignment(phi, theta, bvec, sympy=True):
     )
 
 
-def myelin_phase(shell, phi, theta, mean_log_g, sympy=True):
+def myelin_phase(scanner, b0_dir, t_phase, phi, theta, mean_log_g, sympy=True):
     """
     Computes the intra-axonal phase accumulation due to myelin in rad
 
@@ -758,8 +731,8 @@ def myelin_phase(shell, phi, theta, mean_log_g, sympy=True):
     :param sympy: set to True for sympy input, False for numpy input
     :return: phase accumulated at second readout in rad
     """
-    return -0.75 * shell.scanner.larmor_frequency * mean_log_g * (
-            1 - alignment(phi, theta, shell.b0_dir, sympy) ** 2) * shell.scanner.anisotropic_susceptibility * 1e-9 * shell.t_phase
+    return -0.75 * scanner.larmor_frequency * mean_log_g * (
+            1 - alignment(phi, theta, b0_dir, sympy) ** 2) * scanner.anisotropic_susceptibility * 1e-9 * t_phase
 
 
 def single_fit(parameters: Parameters, data, tofit, model: Model, scipy_kwargs=None, only_mag=False,
@@ -909,11 +882,18 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
         angles = ['theta', 'phi']
     #print("orientation after DTI:", parameters.as_dict['theta'], parameters.as_dict['phi'])
     ase_mags = ['R2', 'R2i'] if model.T2 else ['R2_attenuation']
-    best_fit = single_fit(parameters, data, ['amplitude_se'], model, only_mag=True, verbose=verbose)
-    best_fit = single_fit(best_fit, data, ase_mags, model, only_mag=True, verbose=verbose)
-    best_fit = single_fit(best_fit, data, ['width', 'amplitude_se'] + ase_mags, model, only_mag=True, verbose=verbose)
-    best_fit = single_fit(best_fit, data, angles + ['width', 'amplitude_se'] + ase_mags, model,
-                          only_mag=True, verbose=verbose)
+    best_fit = single_fit(
+        parameters, data, ['amplitude_se'], model, only_mag=True, verbose=verbose,
+        scipy_kwargs={'tol': 1e-12})
+    best_fit = single_fit(
+        best_fit, data, ase_mags, model, only_mag=True, verbose=verbose,
+        scipy_kwargs={'tol': 1e-12})
+    best_fit = single_fit(
+        best_fit, data, ['width', 'amplitude_se'] + ase_mags, model, only_mag=True, verbose=verbose,
+        scipy_kwargs={'tol': 1e-12})
+    best_fit = single_fit(
+        best_fit, data, angles + ['width', 'amplitude_se'] + ase_mags, model, only_mag=True, verbose=verbose,
+        scipy_kwargs={'tol': 1e-12})
 
     # fitting phase as well
     if movement_phase:
@@ -957,7 +937,7 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
                               verbose=verbose)
 
         final_fits.append(single_fit(iter_fit, data, log_g_vars + phase_confounds, model, verbose=verbose,
-                                     scipy_kwargs=dict(options=dict(ftol=1e-9), method='L-BFGS-B')))
+                                     scipy_kwargs=dict(tol=1e-11)))
         ct = model.cost(final_fits[-1], data)
         consistent = np.isclose(ct, best_value, rtol=1e-1)
         if ct < best_value:
@@ -972,6 +952,52 @@ def fit(parameters, data, model: Model, even_eddy=False, only_myelin=False, move
     mp = ["movement_phase", ] if movement_phase else []
     res = single_fit(best_iter_fit, data, angles + ['width', ] + mp +
                         log_g_vars + ase_mags + phase_confounds, model, verbose=verbose,
-                        scipy_kwargs=dict(options=dict(ftol=1e-13), method='L-BFGS-B'))
+                        scipy_kwargs=dict(tol=1e-16))
     res.simplify_params()
     return res
+
+
+def _cost_helper(self, new_params, struct_params, data, all_functions, only_mag, tofit, map_derivative, edges):
+    ct = 0.
+    dp = np.zeros(new_params.size)
+    all_params = self.func_args(struct_params)
+    for idx_shell, shell_data, shell_params, shell in zip(range(len(data)), data, all_params, struct_params.shells):
+        all_evals = []
+        for fibre_params in shell_params:
+            all_evals.append(all_functions(fibre_params))
+
+        for idx_readout, observed in enumerate(shell_data):
+            signal = np.zeros(shell.ngradients, dtype=float if only_mag else complex)
+            for s in all_evals:
+                signal += s['signal'][idx_readout]
+            if only_mag or not np.iscomplex(observed[0]):
+                ct_part = ((np.real(signal) - abs(observed)) ** 2).sum()
+                der_signal = 2 * (np.real(signal) - abs(observed))
+            else:
+                ct_part = (abs(signal - observed) ** 2).sum()
+                der_signal = np.conj(2 * (signal - observed))
+            ct += ct_part
+
+            for idx_fibre, derivatives in enumerate(all_evals):
+                if 'eddy_odd' in tofit or 'eddy_even' in tofit:
+                    dphase = derivatives['eddy_phase'][idx_readout] * der_signal
+                for name in tofit:
+                    if name.startswith('eddy_') and struct_params.eddy_lmax == 0:
+                        der_param_seq = np.zeros((), dtype=complex)
+                    elif name == 'eddy_odd':
+                        der_param_seq = shell.spherical_harmonics(struct_params.eddy_lmax, odd=True).T.dot(dphase.real)
+                    elif name == 'eddy_even':
+                        der_param_seq = shell.spherical_harmonics(struct_params.eddy_lmax, odd=False).T.dot(dphase.real)
+                    else:
+                        der_param_seq = derivatives[name][idx_readout]
+                    if name == 'movement_phase':
+                        der_param = der_param_seq * der_signal
+                    elif name.startswith('eddy_'):
+                        der_param = der_param_seq
+                    elif np.asarray(der_param_seq).ndim == 0:
+                        der_param = der_param_seq * der_signal.sum()
+                    else:
+                        der_param = np.dot(der_param_seq, der_signal)
+                    idx_start = map_derivative[(name, idx_shell, idx_fibre)]
+                    dp[edges[idx_start]:edges[idx_start+1]] += der_param.real
+    return ct, dp
\ No newline at end of file
