@@ -2,6 +2,7 @@
 Defines the acquisition parameters for a shell of DIPPI data
 """
 from enum import Enum
+from functools import lru_cache
 from scipy import optimize
 from dataclasses import dataclass, InitVar, field
 import numpy as np
@@ -135,7 +136,11 @@ class ObservedShell(DataClassJsonMixin):
         return self.TE2 + self.t_phase
 
     def spherical_harmonics(self, lmax, odd=True):
-        return spherical_harmonics(self.gradients, lmax, odd)
+        if not hasattr(self, '_sh_cache'):
+            self._sh_cache = {}
+        if (lmax, odd) not in self._sh_cache:
+            self._sh_cache[(lmax, odd)] = spherical_harmonics(self.gradients, lmax, odd)
+        return self._sh_cache[(lmax, odd)]
 
     @property
     def ngradients(self, ):
